/*************************************************************************
Title:    arthernet external memory test program
Author:   Michael Odenwald
Software: WinAVR 20060421
Hardware: athernet with ATMega128 running at 16MHz
Date:     16.06.2006

*************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "extmem.h"
#include "cpld.h"
#include "uart.h"


/* define CPU frequency in Mhz here if not defined in Makefile */
#ifndef F_CPU
#define F_CPU 16000000UL
#endif

/* baud rate */
#define UART_BAUD_RATE 57600

/* the number of byte for the malloc test */
#define MALLOC_MEM_TEST_SIZE 256

//
// mem_dump
//
// function to print a memory block, at the given address and size
// none prinable chars will be fitt by a dot
//
void mem_dump(void *adr, size_t len)
{
	unsigned char outbuf[64];
	unsigned char c[17];
	uint8_t *start = adr;
	unsigned char i;

	sprintf(outbuf, "Dumping %d bytes:\r\n\r\n", len + 1);
	uart_puts(outbuf);

	sprintf(outbuf, "Address: 00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F\r\n");
	uart_puts(outbuf);

	for(; ((uint8_t *) adr) < ((uint8_t *)(((uint16_t) start) + ((uint16_t) len))); adr += 16)
	{
		sprintf(outbuf, " %p: ", adr);
		uart_puts(outbuf);

		for(i = 0; i < 16; i++)
		{
			if (*((unsigned char *) (adr + i)) > 0x20 &&  *((unsigned char *) (adr + i)) != 0x9b)
			{
				c[i] = *((unsigned char *) (adr + i));
			}
			else
			{
				c[i] = ('.');
			}

			sprintf(outbuf, "%02X ", *((uint8_t *)(adr + i)));	/* does not work with minimalistic printf-variant */

#			if 0
			/* Workaround for minimalistic printf-variant */
			if((*((uint8_t *)(adr + i))) >= 0x10)
				sprintf(outbuf, "%X ", *((uint8_t *)(adr + i)));
			else
				sprintf(outbuf, "0%X ", *((uint8_t *)(adr + i)));
#			endif

			uart_puts(outbuf);
		}

		c[16] = 0;

		sprintf(outbuf, ":%s:\r\n", c);
		uart_puts(outbuf);
	}
	uart_puts("\r\n");
}


//
// the main function
//
int main(void)
{
	volatile unsigned char *pSRAM = NULL;
    unsigned char outbuf[128];
    uint16_t c;
    uint16_t tStart = 0, tStop = 0;
    uint8_t  data;
	uint8_t *pExtMallocBlock = NULL;

	//
	// init hardware URAT and the library
	//
    uart_init( UART_BAUD_SELECT(UART_BAUD_RATE,F_CPU) );

    //
    // now enable interrupt, since UART library is interrupt controlled
    //
    sei();

    // run forever
	while(1)
	{
		sprintf(outbuf, "\n\rWelcome to the arthernet external memory test program\n\r");
		uart_puts(outbuf);

		sprintf(outbuf, "-- System Info --\n\r");
		uart_puts(outbuf);

		sprintf(outbuf, "   Heap-Start: %p; -End: %p\r\n", __malloc_heap_start, __malloc_heap_end);
		uart_puts(outbuf);

		sprintf(outbuf, "   Memory Bank Register: 0x%02x\r\n", *(unsigned char *)pBank_Reg);
		uart_puts(outbuf);

		sprintf(outbuf, "\r\n-- Starting external SRAM test --\r\n");
		uart_puts(outbuf);

		sprintf(outbuf, "   o writing %d bytes ...\r\n", ((uint8_t *) 0xFFFF) - ((uint8_t *) 0x8000));
		uart_puts(outbuf);

		data = 0;

		//
		// fill the memory bank 1 (32kByte) with 0x00 .. 0xff data and than start again
		//
		for(pSRAM = pBank_1; pSRAM < (unsigned char *) 0xFFFF; pSRAM++)
		{
			*pSRAM = data++;
		}

		sprintf(outbuf, "   o comparing ...\r\n");
		uart_puts(outbuf);

		data = 0;

		//
		// test the memory (32kByte) with the writen data 0x00 .. 0xff and so on.
		//
		for(pSRAM = pBank_1; pSRAM < (unsigned char *) 0xFFFF; pSRAM++)
		{
			if(*pSRAM != data)
			{
				sprintf(outbuf, "\r\n   ** @0x%p ERROR; expected: 0x%02X, read: 0x%02X\r\n", pSRAM, data, *pSRAM);
				uart_puts(outbuf);
			}
			data++;
		}

		sprintf(outbuf, "   o Test done.\r\n");
		uart_puts(outbuf);

		sprintf(outbuf, "   o allocating %d bytes with system malloc() ...\r\n", MALLOC_MEM_TEST_SIZE);
		uart_puts(outbuf);

		//
		// measure the CPU ticks for the malloc statement
		//
		TCNT3 = 0x0000;
		TCCR3B = (1 << CS30);

		tStart = TCNT3;
		pExtMallocBlock = malloc(256);
		tStop = TCNT3;

		sprintf(outbuf, "   o fill with test pattern 0xAA ...\r\n");
		uart_puts(outbuf);

		//
		// fill the memory block with the test pattern
		//
		memset(pExtMallocBlock, 0xAA, 256);

		sprintf(outbuf, "   o Test done, malloc() took %d cycles on this system.\r\n", (tStop - tStart));
		uart_puts(outbuf);

		sprintf(outbuf, "-- Dump memory --\r\n\r\n");
		uart_puts(outbuf);

		sprintf(outbuf, "-- Bank 0 (low 32kByte) --\r\n\r\n");
		uart_puts(outbuf);

		mem_dump(pExtMallocBlock, (size_t) 0xFF);
		free(pExtMallocBlock);

		sprintf(outbuf, "-- Bank 1 (high 32kByte) --\r\n\r\n");
		uart_puts(outbuf);

		mem_dump((void *)0x8000, (size_t) 0xFF /*((uint16_t) 0xFFFF - (uint16_t) 0x8000)*/);

		sprintf(outbuf, "-- End of memory dump --\r\n");
		uart_puts(outbuf);

		while(1)
		{
			c = uart_getc();
	        if ( c & UART_NO_DATA )
	        {
	            /*
	             * no data available from UART
	             */
	        }
	        else
	        {
	            /*
	             * new data available from UART
	             * check for Frame or Overrun error
	             */
	            if ( c & UART_FRAME_ERROR )
	            {
	                /* Framing Error detected, i.e no stop bit detected */
	                uart_puts_P("UART Frame Error: ");
	            }
	            if ( c & UART_OVERRUN_ERROR )
	            {
	                /*
	                 * Overrun, a character already present in the UART UDR register was
	                 * not read by the interrupt handler before the next character arrived,
	                 * one or more received characters have been dropped
	                 */
	                uart_puts_P("UART Overrun Error: ");
	            }
	            if ( c & UART_BUFFER_OVERFLOW )
	            {
	                /*
	                 * We are not reading the receive buffer fast enough,
	                 * one or more received character have been dropped
	                 */
	                uart_puts_P("Buffer overflow error: ");
	            }
	            /*
	             * send received character back
	             */
	            uart_putc( (unsigned char)c );
	            break;
			}
		}

	}
}
